from typing import Any, Dict

from .._models import EvaluationResult
from ._evaluator_base import EvaluatorBase


class TrajectoryEvaluator(EvaluatorBase):
    """Evaluator that analyzes the trajectory/path taken to reach outputs."""

    def __init__(
        self,
        trajectory_config: Dict[str, Any],
        step_weights: Dict[str, float],
        target_output_key: str = "*",
    ):
        """Initialize the trajectory evaluator.

        Args:
            trajectory_config: Configuration for trajectory analysis
            step_weights: Weights for different steps in the trajectory
            target_output_key: Key in output to evaluate ("*" for entire output)
        """
        super().__init__()
        self.trajectory_config = trajectory_config or {}
        self.step_weights = step_weights or {}
        self.target_output_key = target_output_key

    async def evaluate(
        self,
        evaluation_id: str,
        evaluation_name: str,
        input_data: Dict[str, Any],
        expected_output: Dict[str, Any],
        actual_output: Dict[str, Any],
    ) -> EvaluationResult:
        """Evaluate using trajectory analysis.

        Args:
            evaluation_id: The ID of the evaluation being processed
            evaluation_name: The name of the evaluation
            input_data: The input data for the evaluation
            expected_output: The expected output
            actual_output: The actual output from the agent

        Returns:
            EvaluationResult containing the score and details
        """
        raise NotImplementedError()



from contextlib import contextmanager
from enum import Enum
from typing import Any, Dict, Iterator, List, Optional, Type, TypeVar

import click
from rich.console import Console
from rich.live import Live
from rich.progress import (
    Progress,
    SpinnerColumn,
    TaskID,
    TextColumn,
    TimeElapsedColumn,
)
from rich.spinner import Spinner as RichSpinner
from rich.text import Text


class LogLevel(Enum):
    """Enum for log levels with corresponding emojis."""

    INFO = ""
    SUCCESS = click.style("✓ ", fg="green", bold=True)
    WARNING = "⚠️"
    ERROR = "❌"
    HINT = "💡"
    CONFIG = "🔧"
    SELECT = "👇"
    LINK = "🔗"
    MAGIC = "✨"


T = TypeVar("T", bound="ConsoleLogger")


class ConsoleLogger:
    """A singleton wrapper class for terminal output with emoji support and spinners."""

    # Class variable to hold the singleton instance
    _instance: Optional["ConsoleLogger"] = None

    def __new__(cls: Type[T]) -> T:
        """Ensure only one instance of ConsoleLogger is created.

        Returns:
            The singleton instance of ConsoleLogger
        """
        if cls._instance is None:
            cls._instance = super(ConsoleLogger, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance  # type: ignore

    def __init__(self):
        """Initialize the ConsoleLogger (only once)."""
        # Only initialize once
        if not getattr(self, "_initialized", False):
            self._console = Console()
            self._spinner_live: Optional[Live] = None
            self._spinner = RichSpinner("dots")
            self._progress: Optional[Progress] = None
            self._progress_tasks: Dict[str, TaskID] = {}
            self._initialized = True

    def _stop_spinner_if_active(self) -> None:
        """Internal method to stop the spinner if it's active."""
        if self._spinner_live and self._spinner_live.is_started:
            self._spinner_live.stop()
            self._spinner_live = None

    def _stop_progress_if_active(self) -> None:
        """Internal method to stop the progress if it's active."""
        if self._progress:
            self._progress.stop()
            self._progress = None
            self._progress_tasks.clear()

    def log(
        self, message: str, level: LogLevel = LogLevel.INFO, fg: Optional[str] = None
    ) -> None:
        """Log a message with the specified level and optional color.

        Args:
            message: The message to log
            level: The log level (determines the emoji)
            fg: Optional foreground color for the message
        """
        # Stop any active spinner before logging
        self._stop_spinner_if_active()

        if not level == LogLevel.INFO:
            emoji = level.value
            if fg:
                formatted_message = f"{emoji} {click.style(message, fg=fg)}"
            else:
                formatted_message = f"{emoji} {message}"
        else:
            formatted_message = message

        click.echo(formatted_message, err=LogLevel.ERROR in (level,))

    def success(self, message: str) -> None:
        """Log a success message."""
        self.log(message, LogLevel.SUCCESS)

    def error(self, message: str, include_traceback: bool = False) -> None:
        """Log an error message with optional traceback.

        Args:
            message: The error message to display
            include_traceback: Whether to include the current exception traceback
        """
        self.log(message, LogLevel.ERROR, "red")

        if include_traceback:
            import traceback

            click.echo(traceback.format_exc(), err=True)

        click.get_current_context().exit(1)

    def warning(self, message: str) -> None:
        """Log a warning message."""
        self.log(message, LogLevel.WARNING, "yellow")

    def info(self, message: str) -> None:
        """Log an informational message."""
        self.log(message, LogLevel.INFO)

    def hint(self, message: str) -> None:
        """Log a hint message."""
        self.log(message, LogLevel.HINT)

    def magic(self, message: str) -> None:
        """Log a magic message."""
        self.log(message, LogLevel.MAGIC, "green")

    def config(self, message: str) -> None:
        """Log a configuration message."""
        self.log(message, LogLevel.CONFIG)

    def select(self, message: str) -> None:
        """Log a selection message."""
        self.log(message, LogLevel.SELECT)

    def link(self, message: str, url: str) -> None:
        """Log a clickable link.

        Args:
            message: The message to display
            url: The URL to link to
        """
        formatted_url = f"\u001b]8;;{url}\u001b\\{url}\u001b]8;;\u001b\\"
        self.log(
            f"{message} {click.style(formatted_url, fg='bright_blue', bold=True)}",
            LogLevel.LINK,
        )

    def prompt(self, message: str, **kwargs: Any) -> Any:
        """Wrapper for click.prompt with emoji.

        Args:
            message: The prompt message
            **kwargs: Additional arguments to pass to click.prompt

        Returns:
            The user's input
        """
        # Stop any active spinner before prompting
        self._stop_spinner_if_active()

        return click.prompt(click.style(f"{message}", fg="yellow", bold=True), **kwargs)

    def display_options(
        self, options: List[Any], message: str = "Select an option:"
    ) -> None:
        """Display a list of options with indices.

        Args:
            options: List of options to display
            message: Optional message to display before the options
        """
        self.select(message)
        for idx, option in enumerate(options, start=0):
            click.echo(f"  {idx}: {option}")

    @contextmanager
    def spinner(self, message: str = "") -> Iterator[None]:
        """Context manager for spinner operations.

        Args:
            message: The message to display alongside the spinner

        Yields:
            None
        """
        try:
            # Stop any existing spinner before starting a new one
            self._stop_spinner_if_active()

            self._spinner.text = Text(message)
            self._spinner_live = Live(
                self._spinner,
                console=self._console,
                refresh_per_second=10,
                transient=False,
                auto_refresh=True,
            )
            self._spinner_live.start()
            yield
        finally:
            self._stop_spinner_if_active()

    def update_spinner(self, message: str) -> None:
        """Update the message of an active spinner.

        Args:
            message: The new message to display
        """
        if self._spinner_live and self._spinner_live.is_started:
            self._spinner.text = Text(message)

    @contextmanager
    def evaluation_progress(
        self, evaluations: List[Dict[str, str]]
    ) -> Iterator["EvaluationProgressManager"]:
        """Context manager for evaluation progress tracking.

        Args:
            evaluations: List of evaluation items with 'id' and 'name' keys

        Yields:
            EvaluationProgressManager instance
        """
        try:
            # Stop any existing progress or spinner
            self._stop_spinner_if_active()
            self._stop_progress_if_active()

            # Create progress with custom columns
            self._progress = Progress(
                SpinnerColumn(),
                TextColumn("[bold blue]{task.description}"),
                TimeElapsedColumn(),
                console=self._console,
                transient=False,
            )

            # Add tasks for each evaluation
            for eval_item in evaluations:
                task_id = self._progress.add_task(eval_item["name"], total=1)
                self._progress_tasks[eval_item["id"]] = task_id

            self._progress.start()

            yield EvaluationProgressManager(self._progress, self._progress_tasks)

        finally:
            self._stop_progress_if_active()

    @classmethod
    def get_instance(cls) -> "ConsoleLogger":
        """Get the singleton instance of ConsoleLogger.

        Returns:
            The singleton instance
        """
        if cls._instance is None:
            return cls()
        return cls._instance


class EvaluationProgressManager:
    """Manager for evaluation progress updates."""

    def __init__(self, progress: Progress, tasks: Dict[str, TaskID]):
        """Initialize the progress manager.

        Args:
            progress: The Rich Progress instance
            tasks: Mapping of evaluation IDs to task IDs
        """
        self.progress = progress
        self.tasks = tasks

    def complete_evaluation(self, eval_id: str) -> None:
        """Mark an evaluation as completed.

        Args:
            eval_id: The evaluation ID
        """
        if eval_id in self.tasks:
            task_id = self.tasks[eval_id]
            # Update description to show completion
            current_desc = self.progress.tasks[task_id].description
            self.progress.update(
                task_id,
                completed=1,
                description=f"[green]✅ {current_desc}[/green]",
            )

    def fail_evaluation(self, eval_id: str, error_message: str) -> None:
        """Mark an evaluation as failed.

        Args:
            eval_id: The evaluation ID
            error_message: The error message
        """
        if eval_id in self.tasks:
            task_id = self.tasks[eval_id]
            # Truncate error message if too long
            short_error = (
                error_message[:40] + "..." if len(error_message) > 40 else error_message
            )
            # Update the description to show failure
            current_desc = self.progress.tasks[task_id].description
            self.progress.update(
                task_id,
                description=f"[red]❌ {current_desc} - {short_error}[/red]",
            )



# type: ignore
import asyncio
import os
from typing import Any
from urllib.parse import urlparse

import click
from dotenv import load_dotenv

from ..telemetry import track
from ._push.sw_file_handler import SwFileHandler
from ._utils._console import ConsoleLogger
from ._utils._constants import (
    UIPATH_PROJECT_ID,
)
from ._utils._project_files import (
    ensure_config_file,
    get_project_config,
    validate_config,
)
from ._utils._uv_helpers import handle_uv_operations

console = ConsoleLogger()
load_dotenv(override=True)


def get_org_scoped_url(base_url: str) -> str:
    parsed = urlparse(base_url)
    org_name, *_ = parsed.path.strip("/").split("/")

    org_scoped_url = f"{parsed.scheme}://{parsed.netloc}/{org_name}"
    return org_scoped_url


async def upload_source_files_to_project(
    project_id: str,
    config_data: dict[Any, str],
    directory: str,
    include_uv_lock: bool = True,
) -> None:
    """Upload source files to UiPath project.

    This function handles the pushing of local files to the remote project:
    - Updates existing files that have changed
    - Uploads new files that don't exist remotely
    - Deletes remote files that no longer exist locally
    - Optionally includes the UV lock file
    """
    sw_file_handler = SwFileHandler(
        project_id=project_id,
        directory=directory,
        include_uv_lock=include_uv_lock,
    )

    await sw_file_handler.upload_source_files(config_data)


@click.command()
@click.argument(
    "root", type=click.Path(exists=True, file_okay=False, dir_okay=True), default="."
)
@click.option(
    "--nolock",
    is_flag=True,
    help="Skip running uv lock and exclude uv.lock from the package",
)
@track
def push(root: str, nolock: bool) -> None:
    """Push local project files to Studio Web Project.

    This command pushes the local project files to a UiPath Studio Web project.
    It ensures that the remote project structure matches the local files by:
    - Updating existing files that have changed
    - Uploading new files
    - Deleting remote files that no longer exist locally
    - Optionally managing the UV lock file

    Args:
        root: The root directory of the project
        nolock: Whether to skip UV lock operations and exclude uv.lock from push

    Environment Variables:
        UIPATH_PROJECT_ID: Required. The ID of the UiPath Cloud project

    Example:
        $ uipath push
        $ uipath push --nolock
    """
    ensure_config_file(root)
    config = get_project_config(root)
    validate_config(config)

    if not os.getenv(UIPATH_PROJECT_ID, False):
        console.error("UIPATH_PROJECT_ID environment variable not found.")

    with console.spinner("Pushing coded UiPath project to Studio Web..."):
        try:
            if not nolock:
                handle_uv_operations(root)

            asyncio.run(
                upload_source_files_to_project(
                    os.getenv(UIPATH_PROJECT_ID),
                    config,
                    root,
                    include_uv_lock=not nolock,
                )
            )
        except Exception as e:
            console.error(f"Failed to push UiPath project: {e}")



import json
from typing import Any, Dict

from ...._config import Config
from ...._execution_context import ExecutionContext
from ...._services.llm_gateway_service import UiPathLlmChatService
from ...._utils.constants import (
    ENV_BASE_URL,
    ENV_UIPATH_ACCESS_TOKEN,
    ENV_UNATTENDED_USER_ACCESS_TOKEN,
    COMMUNITY_agents_SUFFIX,
)
from .._models import EvaluationResult, LLMResponse
from .._models._evaluators import ScoreType
from ._evaluator_base import EvaluatorBase


class LlmAsAJudgeEvaluator(EvaluatorBase):
    """Evaluator that uses an LLM to judge the quality of outputs."""

    def __init__(self, prompt: str = "", model: str = "", target_output_key: str = "*"):
        """Initialize the LLM-as-a-judge evaluator.

        Args:
            prompt: The prompt template for the LLM
            model: The model to use for evaluation
            target_output_key: Key in output to evaluate ("*" for entire output)
        """
        super().__init__()
        self.actual_output_placeholder = "{{ActualOutput}}"
        self.expected_output_placeholder = "{{ExpectedOutput}}"
        self._initialize_llm()
        self.prompt = prompt
        self.model = model
        self.target_output_key: str = target_output_key

    def _initialize_llm(self):
        """Initialize the LLM used for evaluation."""
        import os

        base_url_value: str = os.getenv(ENV_BASE_URL)  # type: ignore
        secret_value: str = os.getenv(ENV_UNATTENDED_USER_ACCESS_TOKEN) or os.getenv(
            ENV_UIPATH_ACCESS_TOKEN
        )  # type: ignore
        config = Config(
            base_url=base_url_value,
            secret=secret_value,
        )
        self.llm = UiPathLlmChatService(config, ExecutionContext())

    async def evaluate(
        self,
        evaluation_id: str,
        evaluation_name: str,
        input_data: Dict[str, Any],
        expected_output: Dict[str, Any],
        actual_output: Dict[str, Any],
    ) -> EvaluationResult:
        """Evaluate using an LLM as a judge.

        Args:
            evaluation_id: The ID of the evaluation being processed
            evaluation_name: The name of the evaluation
            input_data: The input data for the evaluation
            expected_output: The expected output
            actual_output: The actual output from the agent

        Returns:
            EvaluationResult containing the score and details
        """
        # Extract the target value to evaluate
        target_value = self._extract_target_value(actual_output)
        expected_value = self._extract_target_value(expected_output)

        # Create the evaluation prompt
        evaluation_prompt = self._create_evaluation_prompt(expected_value, target_value)

        llm_response = await self._get_llm_response(evaluation_prompt)

        return EvaluationResult(
            evaluation_id=evaluation_id,
            evaluation_name=evaluation_name,
            evaluator_id=self.id,
            evaluator_name=self.name,
            score=llm_response.score,
            input=input_data,
            expected_output=expected_output,
            actual_output=actual_output,
            details=llm_response.justification,
            score_type=ScoreType.NUMERICAL,
        )

    def _extract_target_value(self, output: Dict[str, Any]) -> Any:
        """Extract the target value from output based on target_output_key."""
        if self.target_output_key == "*":
            return output

        # Handle nested keys
        keys = self.target_output_key.split(".")
        value = output

        try:
            for key in keys:
                if isinstance(value, dict):
                    value = value[key]
                else:
                    return None
            return value
        except (KeyError, TypeError):
            return None

    def _create_evaluation_prompt(
        self, expected_output: Any, actual_output: Any
    ) -> str:
        """Create the evaluation prompt for the LLM."""
        formatted_prompt = self.prompt.replace(
            self.actual_output_placeholder,
            str(actual_output),
        )
        formatted_prompt = formatted_prompt.replace(
            self.expected_output_placeholder,
            str(expected_output),
        )

        return formatted_prompt

    async def _get_llm_response(self, evaluation_prompt: str) -> LLMResponse:
        """Get response from the LLM.

        Args:
            evaluation_prompt: The formatted prompt to send to the LLM

        Returns:
            LLMResponse with score and justification
        """
        try:
            # remove community-agents suffix from llm model name
            model = self.model
            if model.endswith(COMMUNITY_agents_SUFFIX):
                model = model.replace(COMMUNITY_agents_SUFFIX, "")

            # Prepare the request
            request_data = {
                "model": model,
                "messages": [{"role": "user", "content": evaluation_prompt}],
                "response_format": {
                    "type": "json_schema",
                    "json_schema": {
                        "name": "evaluation_response",
                        "schema": {
                            "type": "object",
                            "properties": {
                                "score": {
                                    "type": "number",
                                    "minimum": 0,
                                    "maximum": 100,
                                    "description": "Score between 0 and 100",
                                },
                                "justification": {
                                    "type": "string",
                                    "description": "Explanation for the score",
                                },
                            },
                            "required": ["score", "justification"],
                        },
                    },
                },
            }

            response = await self.llm.chat_completions(**request_data)

            try:
                return LLMResponse(**json.loads(response.choices[-1].message.content))
            except (json.JSONDecodeError, ValueError) as e:
                return LLMResponse(
                    score=0.0, justification=f"Error parsing LLM response: {str(e)}"
                )

        except Exception as e:
            # Fallback in case of any errors
            return LLMResponse(
                score=0.0, justification=f"Error during LLM evaluation: {str(e)}"
            )



from uipath._cli._evals._models._evaluation_set import EvaluationItem, EvaluationSet
from uipath._cli._evals._models._evaluators import (
    EvaluationResult,
    EvaluationSetResult,
    EvaluatorCategory,
    EvaluatorType,
    LLMResponse,
)

__all__ = [
    "LLMResponse",
    "EvaluatorCategory",
    "EvaluatorType",
    "EvaluationResult",
    "EvaluationSetResult",
    "EvaluationItem",
    "EvaluationSet",
]
